/*
    ADF-Copy - Reads and writes Amiga floppy disks
    Copyright (C) 2016 Dominik Tonn (nick@niteto.de)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

/*
   encodeMFM is partly based upon floppySectorMfmEncode from FLOPPY.C which is part of the Win Fellow Project
   Authors: Petter Schau, Torsten Enderling (carfesh@gmx.net)
   Copyright (C) 1991, 1992, 1996 Free Software Foundation, Inc
   released under the GNU General Public License
*/

/*
  decodesMFM is partly based upon DecodeSectorData and DecodeLongword functions from AFR.C,
  written by Marco Veneri Copyright (C) 1997 and released as public domain
*/

//#define debug

#ifndef __MK20DX256__
#error Not Teensy 3.2 MCU, make sure you get a version 3.2, because the 3.1 ist not 5V tolerant
#endif

#if F_CPU != 96000000L
#error Please compile Teensy set to 96MHz optimized
#endif

#include <Streaming.h>
#define Version "v0.965"
#define BITBAND_ADDR(addr, bit) (((uint32_t)&(addr) - 0x20000000) * 32 + (bit) * 4 + 0x22000000)

int _dens, _index, _drivesel, _motor, _dir, _step, _writedata, _writeen, _track0, _wprot, _readdata, _side, _diskChange;

uint32_t FTChannelValue, FTStatusControlRegister, FTPinMuxPort;


//default Read Timings (DD Disks)
byte low2 = 30;
byte high2 = 115;
byte high3 = 155;
byte high4 = 255;
byte sectors = 11;



#define timerModeHD 0x08    // TOF=0 TOIE=0 CPWMS=0 CLKS=01 (Sys clock) PS=000 (divide by 1)
#define timerModeDD 0x09    // TOF=0 TOIE=0 CPWMS=0 CLKS=01 (Sys clock) PS=001 (divide by 2)
#define filterSettingDD 0   // 4+4×val clock cycles, 48MHz = 4+4*2 = 32 clock cycles = 0.25us
#define filterSettingHD 0   // 4+4×val clock cycles, 48MHz = 4+4*2 = 32 clock cycles = 0.25us
word timerMode = timerModeDD;
word filterSetting = filterSettingDD;

#define motorMaxTick 5    // Idle Seconds before Motor off
#define maxRetries 6      // maximum retries to read a track
#define transTimeDD 1.96    // timing for write transitions
#define transTimeHD 0.98    // timing for write transitions

#define FLOPPY_GAP_BYTES 1482
//#define gapSize 788
#define streamSizeHD 23*1088+FLOPPY_GAP_BYTES //22 sectors + gap + spare sector
#define streamSizeDD 12*1088+FLOPPY_GAP_BYTES //11 sectors + gap + spare sector
#define writeSizeDD 11*1088+FLOPPY_GAP_BYTES //11 sectors + xx bytes gap
#define writeSizeHD 22*1088+FLOPPY_GAP_BYTES //22 sectors + xx bytes gap
int writeSize = writeSizeDD;
#define MFM_MASK  0x55555555

int streamLen = streamSizeDD;
volatile boolean recordOn = false;
volatile int readPtr;

volatile int sample = 0;
int hist[256];
byte weakTracks[168];
byte trackLog[168];
volatile unsigned int bitCount;
volatile byte sectorCnt;
long errors;
int currentTrack = -1;
int logTrack = -1;
String inputString = "";
volatile int motorTick = 0;
volatile boolean motor = false;
String extError;
String cmd, param;
int retries = maxRetries;
float transitionTime = transTimeDD;
int trackTemp;
int hDet, hDet2;
#define HD 1
#define DD 0
int mode = 0;
boolean autoDensity = true;

struct tfloppyPos
{
  byte track;
  byte side;
  byte dir;
};

struct tfloppyPos floppyPos;

struct SectorTable
{
  unsigned long bytePos;
  byte sector;
};
struct SectorTable sectorTable[25];

struct Sector
{
  byte format_type;       //0
  byte track;             //1
  byte sector;            //2
  byte toGap;             //3
  byte os_recovery[16];   //4
  unsigned long header_chksum;  //20
  unsigned long data_chksum;    //24
  byte data[512];         //28
};

struct Track
{
  byte sector[540];
};

struct Track track[22];

volatile byte bCnt = 0;
volatile unsigned long readBuff;
long zeit;
IntervalTimer motorTimer;
IntervalTimer writeTimer;
long writePtr;
int writeBitCnt;
volatile byte dataByte;
volatile boolean writeActive;
byte * stream;
int * streamBitband;

/*
   waits for and reads command from the serial port
*/
void getCommand()
{
  inputString = "";
  while (!Serial.available());
  while (Serial.available())
  {
    char inChar = (char)Serial.read();
    inputString += inChar;
    if (inChar == '\n') {
      break;
    }
  }
}

/*
   execute command
*/
void doCommand()
{
  inputString.replace((char)10, (char)0);
  inputString.replace((char)13, (char)0);
  cmd = inputString;
  param = inputString;
  if (inputString.indexOf(" ") > 0) {
    cmd.remove(inputString.indexOf(" "), inputString.length());
    param.remove(0, inputString.indexOf(" "));
  } else {
    cmd = inputString;
  }
  if (cmd == "help") {
    Serial << "-----------------------------------------------------------------------------\n";
    Serial << "ADF-Copy " << Version << " - Copyright (C) 2016 Dominik Tonn (nick@niteto.de)\n";
    Serial << "-----------------------------------------------------------------------------\n";
    Serial << "read <n> - read logical track #n\n";
    Serial << "write <n> - write logical track #n\n";
    Serial << "testwrite <n> - write logical track #n filled with 0-255\n";
    Serial << "print - prints amiga track with header\n";
    Serial << "error - returns encoded error in ascii\n";
    Serial << "exterr - returns a detailed error message in ascii\n";
    Serial << "weak - returns retry number for last read in binary format\n";
    Serial << "get <n> - reads track #n silent\n";
    Serial << "put <n> - writes track #n silent\n";
    Serial << "download - download track to pc without headers in binary\n";
    Serial << "upload - upload track to mcu without headers in binary\n";
    Serial << "init - goto track 0\n";
    Serial << "hist - prints histogram of track in ascii\n";
    Serial << "flux - returns histogram of track in binary\n";
    Serial << "index - prints index signal timing in ascii\n";
    Serial << "name - reads track 80 an returns disklabel in ascii\n";
    Serial << "dskcng - returns disk change signal in binary\n";
    Serial << "dens - returns density type of inserted disk in ascii\n";
    Serial << "info - prints state of various floppy signals in ascii\n";
    Serial << "enc - encodes data track into mfm\n";
    Serial << "dec - decodes raw mfm into data track\n";
    Serial << "log - prints logical track / tracknumber extracted from sectors\n";
  }
  if (cmd == "read") {
    Serial.printf("Reading Track %d\n", param.toInt());
    gotoLogicTrack(param.toInt());
    errors = readTrack(false);
    if (errors != -1) {
      Serial << "Sectors found: " << sectorCnt << " Errors found: ";
      Serial.println(errors, BIN);
      Serial << "Track expected: " << param.toInt() << " Track found: " << getTrackInfo() << "\n";
      Serial << "bitCount: " << bitCount << "\n";
      Serial.println("OK");
    } else {
      Serial << "bitCount: " << bitCount << "\n";
      Serial << "Read failed!\n";
    }
  }
  if (cmd == "write") {
    Serial.printf("Writing Track %d\n", param.toInt());
    gotoLogicTrack(param.toInt());
    errors = writeTrack();
    if (errors != -1) {
      Serial.println("OK");
    } else {
      Serial << "Write failed!\n";
    }
  }
  if (cmd == "put") {
    trackTemp = param.toInt();
    floppyTrackMfmEncode(trackTemp, (byte*)track, stream);
    gotoLogicTrack(trackTemp);
    errors = writeTrack();
    if (errors == 0) {
      Serial.println("OK");
    } else {
      Serial.println("ERROR");
    }
  }
  if (cmd == "print") {
    printTrack();
    Serial.println("OK");
  }
  if (cmd == "raw") {
    for (int i = 0; i < streamLen; i++) {
      if ((i % 64 == 0)&&(stream[i]!=0)) {
        Serial.println();
      }
      if (stream[i] != 0) {Serial.print(stream[i], HEX);}
    }
    Serial.println("\nOK");
  }
  if (cmd == "dump") {
    byte tByte = 0;
    for (int i = 0; i < streamLen; i=i+2) {
      if ((i % 128 == 0)&&(tByte!=0)) {
        Serial.println();
      }
      tByte = (stream[i] & 0x55) | ((stream[i + 1] & 0x55) << 1);
      if (tByte != 0) {Serial.print(byte2char(tByte));}
    }
    Serial.println("\nOK");
  }
  if (cmd == "error") {
    Serial.println(errors);
  }
  if (cmd == "exterr") {
    Serial.println(extError);
  }
  if (cmd == "weak") {
    Serial.write(weakTracks[logTrack]);
  }
  if (cmd == "get") {
    gotoLogicTrack(param.toInt());
    readTrack(true);
  }
  if (cmd == "goto") {
    gotoLogicTrack(param.toInt());
  }
  if (cmd == "download") {
    downloadTrack();
  }
  if (cmd == "upload") {
    uploadTrack();
    Serial.println("OK");
  }
  if (cmd == "tw") {
    motorOn();
    for (int i = 0; i < sectors; i++) {

      fillSector(i);
    }
    trackTemp = param.toInt();
    floppyTrackMfmEncode(trackTemp, (byte*)track, stream);
    gotoLogicTrack(trackTemp);
    writeTrack();
    retries = 6;
    delay(500);
    if (readTrack(false) != -1) {
      Serial << "Sectors found: " << sectorCnt << " Errors found: ";
      Serial.println(errors, BIN);
      Serial.println("OK");
    } else {
      Serial << "Read failed!\n";
    }
    retries = maxRetries;
  }
  if (cmd == "hist") {
    analyseHist(false);
    printHist();
    Serial.println("OK");
  }
  if (cmd == "flux") {
    analyseHist(true);
    printFlux();
  }
  if (cmd == "index") {
    Serial << indexTimer() << "ms transitions: " << bitCount << "\n";
    Serial.println("OK");
  }
  if (cmd == "getmode") {
    autoDensity = true;
    motorOn();
    if (mode == HD) {
      Serial.println("HD");
    } else {
      Serial.println("DD");
    }
  }
  if (cmd == "name") {
    Serial.println(getName());
  }
  if (cmd == "dskcng") {
    Serial.println(diskChange());
  }
  if (cmd == "dens") {
    Serial.print("Disk Type: ");
    if (hdDisk()) {
      Serial.println("High Density");
    } else {
      Serial.println("Double Density");
    }
  }
  if (cmd == "setmode") {
    if (param.substring(1) == "hd") {
      autoDensity = false;
      setMode(HD);
      Serial.println("HD");
    } else {
      autoDensity = false;
      setMode(DD);
      Serial.println("DD");
    }
  }
  if (cmd == "info") {
    printStatus();
  }
  if (cmd == "ver") {
    Serial << "ADF-Copy Controller Firmware " << Version;
    if (hDet == 0) Serial.println(" Old Hardware");
    if (hDet == 1) Serial.println(" New Hardware");
    if (hDet == -1) Serial.println(" Unknown Hardware");
  }
  if (cmd == "enc") {
    Serial << "Encoding to MFM\n";
    floppyTrackMfmEncode(0, (byte*)track, stream);
    Serial << "done\n";
  }
  if (cmd == "dec") {
    Serial << "Decode MFM\n";
    decodeTrack(false);
    Serial << "done\n";
  }
  if (cmd == "log") {
    Serial << "Tracking Info\n";
    for (int i = 0; i < 160; i++) {
      if (i != trackLog[i]) {
        Serial << "Track: " << i << " Real Track: " << trackLog[i] << "\n";
      }
    }
  }
  if (cmd == "mfm") {
    Serial.println(mfmByte(param.toInt(), 0), HEX);
  }

  if (cmd == "test") {
    Serial.print("streambitband[0]: ");
    Serial.println((long)&streamBitband[0], HEX);
    Serial.print("streambitband[1]: ");
    Serial.println((long)&streamBitband[1], HEX);
    Serial.print("streambitband[2]: ");
    Serial.println((long)&streamBitband[2], HEX);
    for (int i = 0; i < 8; i++) {
      print_binary(stream[i], 8);
    }
    Serial.println();
    for (int i = 0; i < 64; i++) {
      Serial.print(streamBitband[i]);
    }
    Serial.println();
    int zeit = millis();
#define loops 100000
    for (int i = 0; i < loops; i++) {
      asm volatile("nop\n");
    }
    zeit = millis() - zeit;
    zeit = zeit * 1000000 / loops;
    Serial << "Time taken: " << zeit << "ns\n";
    for (int i = 0; i < 256; i++) {
      hist[i] = 0;
    }
    for (int i = 0; i < (writeSize * 8) + 8; i++) {
      hist[getByte(i)]++;
    }
    printHist();
  }


  if (cmd == "init") {
    autoDensity = true;
    currentTrack = -1;
    if (diskChange() == 1) {
      motorOn();
      seek0();
    }
    errors = 0;
  }
}

void setup() {
  hDet = hardwareVersion();
  registerSetup(hDet);
  //  hDet = detectHardware(4);
  //  hDet2 = detectHardware(16);
  pinMode(_dens, OUTPUT);
  digitalWriteFast(_dens, HIGH);
  pinMode(_index, INPUT_PULLUP);
  pinMode(_motor, OUTPUT);
  digitalWriteFast(_motor, HIGH);
  pinMode(_drivesel, OUTPUT);
  digitalWriteFast(_drivesel, HIGH);
  pinMode(_dir, OUTPUT);
  digitalWriteFast(_dir, HIGH);
  pinMode(_step, OUTPUT);
  digitalWriteFast(_step, HIGH);
  pinModeFast(_writedata, OUTPUT);
  digitalWriteFast(_writedata, HIGH);
  pinMode(_writeen, OUTPUT);
  digitalWriteFast(_writeen, HIGH);
  pinMode(_track0, INPUT_PULLUP);
  pinMode(_wprot, INPUT_PULLUP);
  pinMode(_side, OUTPUT);
  digitalWriteFast(_side, HIGH);
  pinMode(_diskChange, INPUT_PULLUP);

  pinMode(13, OUTPUT);
  attachInterrupt(_diskChange, diskChangeIRQ, FALLING);

  motorTimer.priority(255);
  motorTimer.begin(motorTimeout, 1000000);
  Serial.begin(250000);
  while (!Serial);
  inputString.reserve(200);
  extError.reserve(31);
  cmd.reserve(20);
  param.reserve(20);
  Serial << "---------------------------------------------------------------------------------------\n";
  Serial << "ADF-Copy " << Version << " " << __DATE__ << " - Copyright (C) 2016 Dominik Tonn (nick@niteto.de)\n";
  Serial << "released under GNU General Public License, for details see http://www.gnu.org/licenses/\n";
  Serial << "---------------------------------------------------------------------------------------\n";
  Serial << "type help for commands\n";
  byte * tempMem;
  tempMem = (byte*) malloc(1);

  long foo = 0x20000000 - (long)tempMem;
  free(tempMem);
  tempMem = (byte*) malloc(foo);

  // workaround to prevent the compiler from optimizing away the malloc(foo)
  // for aligning the stream memory to bitband memory space
  // you might get a warning that preventCompilerOptimize is never used.
  volatile long preventCompilerOptimize;
  preventCompilerOptimize = (long)&tempMem[foo];
  // end of workaround

  stream = (byte*) malloc(streamSizeHD + 10);
  streamBitband = (int*) BITBAND_ADDR(*stream, 0);
#ifdef debug
  printPtr(tempMem, "tempMem");
  printPtr(stream, "stream");
#endif

  if (stream == 0) {
    Serial.println("Out of memory");
    while (1);
  }
  free(tempMem);
  initRead();
  floppyPos.dir = 0;
  floppyPos.side = 0;
  floppyPos.track = 0;
  initLed();
}
#ifdef debug
void printPtr(byte * dest, String pName)
{
  Serial << pName << ":";
  Serial.print((long)dest, HEX);
  Serial << " -> ";
  Serial.println((long)*dest, HEX);
}
#endif

void loop() {
  getCommand();
  doCommand();
}
